package arithmetic;

import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.Scanner;


public class Operation
{
    private List<StackNode> opStack;
    private double finalResult;
    private boolean assigned;
    
    public Operation(List<StackNode> opStack)
    {
        this.opStack = opStack;    
    }
    
    public static Operation fromString(String operation)
    {
        List<StackNode> stack = new LinkedList<>();
        Scanner s = new Scanner(operation);
        s = s.useDelimiter("\\s+");
        
        while (s.hasNext())
        {
            String element = s.next();
            StackNode node;
            switch (element)
            {
                case "+":
                    node = Operator.ADDICTION;
                    break;
                case "-":
                    node = Operator.SUBTRACTION;
                    break;
                case "*":
                    node = Operator.MULTIPLICATION;
                    break;
                case "/":
                    node = Operator.DIVISION;
                    break;
                default: // mi aspetto un numero
                    try
                    {
                        node = new Number(Double.parseDouble(element));
                    }    
                    catch (NumberFormatException nfe)
                    {
                        continue; // e' qualcosa di sconosciuto: lo ignoro
                    }
                    break;
            }
            stack.add(node);
        }
        return new Operation(stack);    
    }
    
    public synchronized double resolve() // piu' esecutori potrebbero invocare resolve()
    {       
        assert !opStack.isEmpty() || finalResult != 0.0; // se lo stack e' ben formato questo non dovrebbe essere mai vero
        
        // SOLO PER PROVA //
        try // dorme un po'....
        {
            Random rnd = new Random();
            Thread.sleep(500 + rnd.nextInt(1000));
        }
        catch (InterruptedException ie)
        {
            ie.printStackTrace();
        }
        // SOLO PER PROVA //
        
        if (assigned)
        {
            return finalResult; // questa procedura ricorsiva consuma lo stack quindi e' possibile ottenere il risultato una sola volta.
                                // -> lo memorizzo in una variabile e dalla seconda chiamata in poi restituisco quel valore.
        }
        
        double result = 0.0;
        
        StackNode node = opStack.remove(opStack.size() - 1);
        if(node instanceof Operator)
        {
            switch ((Operator)node)
            {
                case ADDICTION:
                    result = resolve() + resolve();
                    break;
                case SUBTRACTION:
                    result = resolve() - resolve();
                    break;
                case MULTIPLICATION:
                    result = resolve() * resolve();
                    break;
                case DIVISION:
                    result = resolve() / resolve();
                    break;
            }
        }
        else if (node instanceof Number) // e' un numero
        {
            result = ((Number)node).getValue();
        }
        
        if (opStack.isEmpty())
        {
            finalResult = result; // se sono finiti i conti, salvo il valore in finalResult
            assigned = true;
        }    
        return result;
    }
    
    // test
    public static void main(String[] args)
    {
        String s1 = "3 4 5 * +"; // -> 23
        String s2 = " 3 4 + 5 *"; // -> 35
        String s3 = "3 4 5 + *"; // -> 27
        
        Operation op = Operation.fromString(s1);
        System.out.println("Risultato 1: " + op.resolve()); 
    
        op = Operation.fromString(s2);
        System.out.println("Risultato 2: " + op.resolve());
        
        op = Operation.fromString(s3);
        System.out.println("Risultato 3: " + op.resolve());
        System.out.println("Risultato 3: " + op.resolve()); // chiamo una seconda volta per vedere se funziona la memorizzazione di variabile.
    }
}