package buffers;

import java.util.concurrent.locks.*;

public class BufferLock<T>
{
	T[] storage;
	int head, tail, size;
	Lock l;
	Condition pieno, vuoto;

	public BufferLock(int max)
	{
		storage = (T[])(new Object[max]);
		head = tail = size = 0;
		l = new ReentrantLock();
		pieno = l.newCondition();
		vuoto = l.newCondition();
	}

	public BufferLock()
	{
		this(10);
	}

	public void insert(T elem)
	{
		try
		{
			l.lock();
		
			while(size >= storage.length)//se avessi la certezza di essere l'unico ad inserire
																 	//potrei usare una condizione non iterata (if)
			{
				//devo aspettare, buffer pieno
				pieno.awaitUninterruptibly();
			}
	
			storage[tail] = elem;
			tail = (tail+1)%storage.length;
			size++;
			if(size == 1)
			{
				vuoto.signal();
			}
		}
		finally
		{
			l.unlock();
		}
	}

	public T remove()
	{
		T retval = null;
		try
		{
			l.lock();
			while(size <= 0)//vedi sopra per la possibilita' di non iterare il test
			{
				//devo aspettare, buffer vuoto
				vuoto.awaitUninterruptibly();
			}
	
			retval = storage[head];
			head = (head + 1)%storage.length;
			if(size == storage.length)
			{
				pieno.signal();
			}
			size--;
		}
		finally
		{
			l.unlock();
		}
		return retval;
	}

	public int size()
	{
		int retval;
		l.lock();
		retval = size;
		l.unlock();
		return retval;
	}
}
