package net;

import java.io.*;
import java.net.*;
import java.util.*;
import remotizzazione.Operazione;

public class ServerTCP5
implements Runnable
{
	ServerSocket ss;
	List<Socket> list;
	List<Box> coda;

	public ServerTCP5(int port) throws IOException
	{
		ss = new ServerSocket(port);
		list = new LinkedList<Socket>();
		coda = new LinkedList<Box>();
		for(int i = 0;i<5;i++)
		{
			(new Thread(new Servant(coda))).start();
		}
	}

	public ServerTCP5() throws IOException
	{
		this(1234);
	}

	public void exec() throws IOException
	{
		Socket s = null;
		while(true)
		{
			s = ss.accept();
			synchronized(list)
			{
				list.add(s);
			}
			(new Thread(this)).start();
		}
	}

	public void run()
	{
		Socket s = null;
		ObjectInputStream ois = null;
		ObjectOutputStream oos = null;
		synchronized(list)
		{
			s = list.remove(0);
		}

		try
		{
			Object o = null;
			oos = new ObjectOutputStream(s.getOutputStream());
			ois = new ObjectInputStream(s.getInputStream());
			while( (o = ois.readObject()) != null)
			{
				//String line = (String)o;
				System.out.println("Ho letto <" + o + "> eseguo");
				Busta bu = (Busta)o;

				Operazione op = (Operazione)bu.payload;
				Box b = new Box();
				b.op = op;
				b.oos = oos;
				b.id = bu.id;
				synchronized(coda)
				{
					coda.add(b);
					coda.notify();
				}
			}
		}
		catch(IOException ioe)
		{
			ioe.printStackTrace();
		}
		catch(ClassNotFoundException cnfe)
		{
			cnfe.printStackTrace();
		}
		finally
		{
			try
			{
				ois.close();
				oos.close();
				s.close();
			}
			catch(IOException ioe2)
			{
				ioe2.printStackTrace();
				System.err.println("Piu' storta di cosi', neanche riesco a chiudere i canali...");
			}
		}
		System.out.println(Thread.currentThread().getName() + " adieu");
	}

	public static void main(String[] argv) throws IOException
	{
		ServerTCP5 serv = null;
		if(argv.length > 0)
		{
			serv = new ServerTCP5(Integer.parseInt(argv[0]));
		}
		else
		{
			serv = new ServerTCP5();
		}
		serv.exec();
	}

	class Servant
	implements Runnable
	{
		List<Box> coda;

		Servant(List<Box> l)
		{
			coda = l;
		}

		public void run()
		{
			System.out.println("Servant " + Thread.currentThread().getName() + " partito, attendo qualcosa da fare");
			while(true)
			{
				Box b = null;
				synchronized(coda)
				{
					while(coda.isEmpty())
					{
						try
						{
							coda.wait();
						}
						catch(InterruptedException ie)
						{
							ie.printStackTrace();
						}
					}
					b = coda.remove(0);
				}
				switch(b.op.operazione)
				{
					case 0:
						b.op.result = b.op.d1 + b.op.d2;
						sleep(1000);
						break;
					case 1:
						b.op.result = b.op.d1 - b.op.d2;
						sleep(2000);
						break;
					case 2:
						b.op.result = b.op.d1 * b.op.d2;
						sleep(3000);
						break;
					case 3:
						b.op.result = b.op.d1 / b.op.d2;
						sleep(4000);
						break;
					default:
						System.err.println("Operazione " + b.op.operazione + " non nota");
				}
				try
				{
					synchronized(b.oos)
					{
						Busta bu = new Busta();
						bu.payload = b.op;
						bu.id = b.id;
						b.oos.writeObject(bu);
						b.oos.flush();
					}
				}
				catch(IOException ioe)
				{
					ioe.printStackTrace();
					System.err.println("Thread " + Thread.currentThread().getName() + " I/O error -> termino");
					return;
				}
			}
		}

		void sleep(long t)
		{
			try
			{
				Thread.sleep(t);
			}
			catch(InterruptedException ie)
			{
				ie.printStackTrace();
			}
		}
	}

	class Box
	{
		Operazione op;
		ObjectOutputStream oos;
		long id;
	}
}
